@@ -36,8 +36,11 @@
 static
 void  altmap_CB(Widget w,XtPointer closure,XtPointer call_data)
 {
+   NewMode=Mode;
    ToLatKeys.Fire();
-   run_xmodmap((const char*)closure);
+   FixNewMode();
+   SwitchKeyboard(NewMode);
+   load_map_delayed((const char*)closure);
 }
 
 Widget   CreateAltMapMenu(Widget parent,const char *altmap)
@@ -56,7 +59,7 @@ Widget   CreateAltMapMenu(Widget parent,const char *altmap)
    for(;;)
    {
       /* ignore leading spaces */
-      while(*scan && isspace(*scan))
+      while(*scan && isspace((unsigned char)*scan))
          scan++;
       if(!*scan)
          break;
@@ -64,7 +67,7 @@ Widget   CreateAltMapMenu(Widget parent,const char *altmap)
       file=scan;
 
       /* track down the end of file name */
-      while(*scan && !isspace(*scan))
+      while(*scan && !isspace((unsigned char)*scan))
          scan++;
 
       if(*scan)
