@@ -1,11 +1,11 @@
 /*
   xrus - keyboard switcher/indicator
-  Copyright (c) 1996 Alexander V. Lukyanov
+  Copyright (c) 1995-1998 Alexander V. Lukyanov
   This is free software with no warranty.
   See COPYING for details.
 
   Athena version of properties dialog
-  by Vladimir Grachev, e-mail vgrachev@ssau.ru
+  by Vladimir Grachev, e-mail vgrachev@ssau.ru (1998)
 */
 /*__________________________________________________________________________
 **
@@ -62,6 +62,7 @@ struct
    Widget   form2,form4;
    Widget   locker,switch_keys,locker_def,switch_keys_def;
    Widget   to_lat_keys,to_rus_keys;
+   Widget   switch_for_one_keys;
    Widget   locker_test,save_button;
 }
    props;
@@ -80,6 +81,7 @@ struct {
   Widget switch_label, switch_keys;
   Widget lat_label, to_lat_keys;
   Widget rus_label, to_rus_keys;
+  Widget switch_for_one_keys;
   Widget default2_button;
   Widget ok_button, apply_button, save_button, cancel_button;
 } props;
@@ -112,6 +114,28 @@ XtArgVal *l_top;
 };
 #endif /* TK==TK_XAW */
 
+static
+char *get_string(Widget w)
+{
+   char  *value;
+#if TK==TK_MOTIF
+   XtVaGetValues(w,XmNvalue,&value,NULL);
+#elif TK==TK_XAW
+   XtVaGetValues(w,XtNstring,&value,NULL);
+#endif
+   return value;
+}
+
+static
+void set_string(Widget w,const char *value)
+{
+#if TK==TK_MOTIF
+   XtVaSetValues(w,XmNvalue,value,NULL);
+#elif TK==TK_XAW
+   XtVaSetValues(w,XtNstring,value,NULL);
+#endif
+}
+
 static
 void  PropsSet()
 {
@@ -119,12 +143,7 @@ void  PropsSet()
    XtVaSetValues(props.allow_locker,XmNset,!!AppData.autolock,NULL);
    XtVaSetValues(props.autolock_delay,XmNvalue,AppData.timeout,NULL);
    XtVaSetValues(props.use_bell,XmNset,!!AppData.useBell,NULL);
-   XtVaSetValues(props.locker,XmNvalue,AppData.locker,NULL);
-   XtVaSetValues(props.switch_keys,XmNvalue,AppData.switchKeys,NULL);
-   XtVaSetValues(props.to_lat_keys,XmNvalue,AppData.toLatKeys,NULL);
-   XtVaSetValues(props.to_rus_keys,XmNvalue,AppData.toRusKeys,NULL);
 #elif TK==TK_XAW
-
    XtVaSetValues(props.locker_toggle,
      XtNstate, !!AppData.autolock,
      NULL);
@@ -138,19 +157,13 @@ void  PropsSet()
    XtVaSetValues(props.bell_toggle,
      XtNstate, !!AppData.useBell,
      NULL);
-   XtVaSetValues(props.locker,
-     XtNstring, AppData.locker,
-     NULL);
-   XtVaSetValues(props.switch_keys,
-     XtNstring, AppData.switchKeys,
-     NULL);
-   XtVaSetValues(props.to_lat_keys,
-     XtNstring, AppData.toLatKeys,
-     NULL);
-   XtVaSetValues(props.to_rus_keys,
-     XtNstring, AppData.toRusKeys,
-     NULL);
 #endif
+
+   set_string(props.locker,AppData.locker);
+   set_string(props.switch_keys,AppData.switchKeys);
+   set_string(props.to_lat_keys,AppData.toLatKeys);
+   set_string(props.to_rus_keys,AppData.toRusKeys);
+   set_string(props.switch_for_one_keys,AppData.switchForOneKeys);
 }
 
 static
@@ -159,18 +172,6 @@ void  ErrorDismiss_CB(Widget w,XtPointer closure,XtPointer call_data)
    XtDestroyWidget(w);
 }
 
-static
-char *get_string(Widget w)
-{
-   char  *value;
-#if TK==TK_MOTIF
-   XtVaGetValues(w,XmNvalue,&value,NULL);
-#elif TK==TK_XAW
-   XtVaGetValues(w,XtNstring,&value,NULL);
-#endif
-   return value;
-}
-
 static
 int  PropsApply()
 {
@@ -213,6 +214,17 @@ int  PropsApply()
       ParseKeyCombination(AppData.toRusKeys,&ToRusKeys,message);
    }
 
+   strcpy(old_switchKeys,AppData.switchForOneKeys);
+   strcpy(AppData.switchForOneKeys,get_string(props.switch_for_one_keys));
+   ParseKeyCombination(AppData.switchForOneKeys,&SwitchForOneKeys,message);
+   if(*message)
+   {
+      strcat(large_message,message);
+
+      strcpy(AppData.switchForOneKeys,old_switchKeys);
+      ParseKeyCombination(AppData.switchForOneKeys,&SwitchForOneKeys,message);
+   }
+
    if(large_message[0])
    {
 #if TK==TK_MOTIF
@@ -308,7 +320,8 @@ void  PropsSave()
       "Xrus*useBell:\t\t%s\n"
       "Xrus*switchKeys:\t%s\n"
       "Xrus*toLatKeys:\t\t%s\n"
-      "Xrus*toRusKeys:\t\t%s\n",
+      "Xrus*toRusKeys:\t\t%s\n"
+      "Xrus*switchForOneKeys:\t%s\n",
 
       AppData.locker,
       AppData.timeout,
@@ -316,7 +329,8 @@ void  PropsSave()
       AppData.useBell ?"true":"false",
       AppData.switchKeys,
       AppData.toLatKeys,
-      AppData.toRusKeys
+      AppData.toRusKeys,
+      AppData.switchForOneKeys
       );
 
    fclose(f);
@@ -388,23 +402,15 @@ void  Default_CB(Widget w,XtPointer closure,XtPointer cd_data)
       /* set switch keys default */
 #if TK==TK_MOTIF
       XtVaSetValues(props.use_bell,XmNset,0,NULL);
-      XtVaSetValues(props.switch_keys,XmNvalue,DefaultSwitchKeys,NULL);
-      XtVaSetValues(props.to_lat_keys,XmNvalue,DefaultToLatKeys,NULL);
-      XtVaSetValues(props.to_rus_keys,XmNvalue,DefaultToRusKeys,NULL);
 #elif TK==TK_XAW
       XtVaSetValues(props.bell_toggle,
         XtNstate, 0,
         NULL);
-      XtVaSetValues(props.switch_keys,
-        XtNstring, DefaultSwitchKeys,
-        NULL);
-      XtVaSetValues(props.to_lat_keys,
-        XtNstring, DefaultToLatKeys,
-        NULL);
-      XtVaSetValues(props.to_rus_keys,
-        XtNstring, DefaultToRusKeys,
-        NULL);
 #endif
+      set_string(props.switch_keys,DefaultSwitchKeys);
+      set_string(props.to_lat_keys,DefaultToLatKeys);
+      set_string(props.to_rus_keys,DefaultToRusKeys);
+      set_string(props.switch_for_one_keys,DefaultSwitchForOneKeys);
       break;
    }
 }
@@ -433,7 +439,7 @@ static void  cb_jump(Widget w, XtPointer client_data, XtPointer percent_ptr) {
 static
 void  PropsInit()
 {
-   static inited=0;
+   static int inited=0;
 
 #if TK==TK_MOTIF
    int   count;
@@ -480,6 +486,8 @@ void  PropsInit()
    props.to_lat_keys=XtVaCreateManagedWidget("to_lat_keys",xmTextFieldWidgetClass,props.form2,NULL);
    XtVaCreateManagedWidget("to_rus_keys_label",xmLabelGadgetClass,props.form2,NULL);
    props.to_rus_keys=XtVaCreateManagedWidget("to_rus_keys",xmTextFieldWidgetClass,props.form2,NULL);
+   XtVaCreateManagedWidget("switch_for_one_keys_label",xmLabelGadgetClass,props.form2,NULL);
+   props.switch_for_one_keys=XtVaCreateManagedWidget("switch_for_one_keys",xmTextFieldWidgetClass,props.form2,NULL);
 
    props.switch_keys_def=XtVaCreateManagedWidget("switch_keys_def",xmPushButtonWidgetClass,props.form2,NULL);
    XtAddCallback(props.switch_keys_def,XmNactivateCallback,Default_CB,(XtPointer)1);
@@ -594,6 +602,16 @@ void  PropsInit()
     XtNeditType, XawtextEdit,
     NULL);
 
+  XtVaCreateManagedWidget("switch_for_one_keys_label",
+    labelWidgetClass, props.form,
+    NULL);
+
+  props.switch_for_one_keys = XtVaCreateManagedWidget("switch_for_one_keys",
+    asciiTextWidgetClass, props.form,
+    XtNtype, XawAsciiString,
+    XtNeditType, XawtextEdit,
+    NULL);
+
   props.default2_button = XtVaCreateManagedWidget("default2_button",
     commandWidgetClass, props.form,
     NULL);
