@@ -1,6 +1,6 @@
 /*
-  xrus - keyboard switcher/indicator
-  Copyright (c) 1995-1998 Alexander V. Lukyanov
+  xrus - keyboard switcher/indicator and autolock.
+  Copyright (c) 1995-2000 Alexander V. Lukyanov
   This is free software with no warranty.
   See COPYING for details.
 */
@@ -31,6 +31,7 @@
 #include <X11/StringDefs.h>
 #include <X11/Shell.h>
 #include <X11/Xmd.h>
+#include <X11/Xlibint.h>   /* for resource_mask */
 
 #if TK==TK_MOTIF
 #include <Xm/Xm.h>
@@ -146,7 +147,7 @@ XtResource        resources[]=
          XtRImmediate,  (XtPointer)False        },
    {  "alwaysMapped","AutoMap",XtRBoolean,sizeof(Boolean),
          XtOffsetOf(XrusRec,alwaysMapped),
-         XtRImmediate,  (XtPointer)True         },
+         XtRImmediate,  (XtPointer)False	},
    {  "capsLockEmu","CapsLockEmu",XtRBoolean,sizeof(Boolean),
          XtOffsetOf(XrusRec,capsLockEmu),
          XtRImmediate,  (XtPointer)False        },
@@ -167,7 +168,25 @@ XtResource        resources[]=
             XtRImmediate,  (XtPointer)False     },
    {  "xmodmapProgram", "XmodmapProgram", XtRString, sizeof(String),
             XtOffsetOf(XrusRec,xmodmap_program),
-            XtRString,  BINDIR "/xrusmodmap"     },
+            XtRString,  BINDIR "/xrusmodmap"    },
+   {  "title0",	  "Title0",   XtRString,  sizeof(String),
+            XtOffsetOf(XrusRec,title0),
+            XtRString,  (XtPointer)0		},
+   {  "title1",	  "Title1",   XtRString,  sizeof(String),
+            XtOffsetOf(XrusRec,title1),
+            XtRString,  (XtPointer)0		},
+   {  "perWindow","PerWindow",XtRBoolean, sizeof(Boolean),
+            XtOffsetOf(XrusRec,per_window_state),
+            XtRImmediate,  (XtPointer)False     },
+   {  "titlePerWindow0","TitlePerWindow0",XtRString,sizeof(String),
+            XtOffsetOf(XrusRec,titlePerWindow0),
+            XtRString,  (XtPointer)0		},
+   {  "titlePerWindow1","TitlePerWindow1",XtRString,sizeof(String),
+            XtOffsetOf(XrusRec,titlePerWindow1),
+            XtRString,  (XtPointer)0		},
+   {  "occupyAllDesks","OccupyAllDesks",XtRBoolean, sizeof(Boolean),
+            XtOffsetOf(XrusRec,occupyAllDesks),
+            XtRImmediate,  (XtPointer)True      },
 };
 
 XrmOptionDescRec  options[]=
@@ -190,6 +209,7 @@ XrmOptionDescRec  options[]=
    {  "+fork",      "*noFork",     XrmoptionNoArg,   "true" },
    {  "-wmicon",    "*wmIcon",     XrmoptionNoArg,   "true"},
    {  "-wmaker",    "*wMaker",     XrmoptionNoArg,   "true"},
+   {  "-perwindow", "*perWindow",  XrmoptionNoArg,   "true"},
 };
 
 Display        *disp;
@@ -215,11 +235,13 @@ enum
    TEMP_LAT=2,
    CAPSLOCK_ON=4,
    FOR_ONE=8,
+   SAVE_MASK=MODE
 };
 int   Mode=LAT;
 int   NewMode;
 
 Widget   top_level=NULL;
+Window   our_host_window=0;
 #if TK!=TK_NONE
 Widget   form_w;
 Widget   switch_button[2];
@@ -230,9 +252,15 @@ int   format_ret;
 unsigned long nitems_ret,bytes_after_ret;
 unsigned char *prop;
 
-int   xrus_check=1;
+static Atom kbd_state_atom;
+static Window focus_window;
+static int    revert_to=RevertToNone;
+static Window last_top_level_focus_window;
+static int    last_top_level_revert_to=RevertToNone;
+static Bool   pushed=False;
+
+static int xrus_check=1;
 
-void  AddChildren(Window);
 void  SetAlarm();
 
 int   (*old_error_handler)(Display *d,XErrorEvent *ev);
@@ -241,20 +269,24 @@ pid_t LockerRunning=0;
 
 time_t alarm_expected;
 
-int   MappingNotifyIgnoreCount=0;
-XtIntervalId   MappingNotifyTimeout=-1;
-XtIntervalId   raise_tmout=-1;
+static int MappingNotifyIgnoreCount=0;
+static XtIntervalId MappingNotifyTimeout=-1;
 
 #if TK!=TK_NONE
-static void  ShowSwitchButton(void);
+static void ShowSwitchButton(void);
+static void SetTitle(int);
 #else
 # define ShowSwitchButton()
+# define SetTitle(x)
 #endif
 
 #define MaxLetterKeySym 0x1000
 #define NationalKeySym(ks) (ks>=128 && ks<MaxLetterKeySym)
 #define AsciiKeySym(ks) (ks>0 && ks<128)
 
+XID client_id_mask;
+#define ClientId(w) ((w)&client_id_mask)
+
 /*________________________________________________________________________
 */
 
@@ -309,6 +341,120 @@ void  FixNewMode()
       NewMode|=TEMP_LAT;
 }
 
+static void SetTitleIndicator(Window w,int mode)
+{
+   XTextProperty prop;
+   char *new_title, *append, *old;
+   int old_len, append_len, title_len;
+
+   if((!AppData.titlePerWindow0    || !AppData.titlePerWindow1)
+   || (!AppData.titlePerWindow0[0] && !AppData.titlePerWindow1[0]))
+      return;
+
+   if(!XGetWMName(disp,w,&prop) || prop.format!=8)
+      return;
+
+   mode&=MODE;
+
+   append=mode?AppData.titlePerWindow1:AppData.titlePerWindow0;
+   old   =mode?AppData.titlePerWindow0:AppData.titlePerWindow1;
+   new_title=alloca(prop.nitems+256);
+   old_len=strlen(old);
+   append_len=strlen(append);
+
+   memcpy(new_title,prop.value,title_len=prop.nitems);
+   new_title[title_len]=0; /* for strcmp below */
+   XFree(prop.value);
+
+   if(title_len>old_len
+   && !strcmp(new_title+title_len-old_len,old))
+      title_len-=old_len;
+   else if(title_len>append_len
+   && !strcmp(new_title+title_len-append_len,append))
+      return;
+
+   if(title_len==0)
+   {
+      strcpy(new_title,"untitled");
+      title_len=strlen(new_title);
+   }
+   if(new_title[title_len-1]!=' ')
+      new_title[title_len++]=' ';
+   strcpy(new_title+title_len,append);
+   title_len+=append_len;
+
+   prop.nitems=title_len;
+   prop.value=new_title;
+   XSetWMName(disp,w,&prop);
+}
+
+static int GetKbdState(Window w)
+{
+   long *prop=0;
+   if(XGetWindowProperty(disp,w,kbd_state_atom,0L,256L,0,XA_CARDINAL,
+         &type_ret,&format_ret,&nitems_ret,&bytes_after_ret,(void*)&prop)
+        ==Success && type_ret==XA_CARDINAL && nitems_ret==1 && prop)
+   {
+      int ret=(prop[0]&SAVE_MASK);
+      XFree(prop);
+      return ret;
+   }
+   return -1;
+}
+
+static void SetTitleIndicatorOnTimeout(XtPointer closure,XtIntervalId *id)
+{
+   Window w=(Window)closure;
+   int mode=GetKbdState(w);
+   if(mode!=-1)
+      SetTitleIndicator(w,mode);
+}
+
+static Bool window_is_top_level(Window w)
+{
+   int num,i;
+   Atom *list=XListProperties(disp,w,&num);
+
+   if(!list)
+      return False;
+
+   for(i=0; i<num; i++)
+   {
+      if(list[i]==XA_WM_NAME
+      || list[i]==XA_WM_CLASS
+      || list[i]==XA_WM_COMMAND)
+      {
+         XFree(list);
+         return True;
+      }
+   }
+   XFree(list);
+   return False;
+}
+
+static Bool window_is_ours(Window w)
+{
+   if(!top_level)
+      return False;
+   return (ClientId(w)==ClientId(XtWindow(top_level)));
+   /* Also possibe:
+   return (XtWindowToWidget(disp,w)!=NULL); */
+}
+
+static void SaveModeForWindow(Window w)
+{
+   long m=Mode&SAVE_MASK;
+   if(!AppData.per_window_state)
+      return;
+   if(w==0 || !window_is_top_level(w))
+      w=last_top_level_focus_window;
+   if(w==0)
+      return;
+   SetTitleIndicator(w,Mode);
+   XChangeProperty(disp,w,kbd_state_atom,XA_CARDINAL,32,
+                   PropModeReplace,(void*)&m,1);
+}
+
 void  SwitchKeyboard(int to)
 {
    KeySym   tmp;
@@ -317,6 +463,7 @@ void  SwitchKeyboard(int to)
    XKeyboardControl  kbd_control;
    int   from_mode,to_mode,from_caps,to_caps;
    int   changed=0;
+   int   from=Mode;
 
    if(to==Mode)
       return;
@@ -401,9 +548,35 @@ void  SwitchKeyboard(int to)
          XBell(disp,30);
    }
    Mode=to;
+   if((from&SAVE_MASK) != (to&SAVE_MASK))
+      SaveModeForWindow(focus_window);
    ShowSwitchButton();
 }
 
+static void SwitchKeyboardForWindow(Window w)
+{
+   int mode;
+   Window old_last_top_level_focus_window=last_top_level_focus_window;
+   Window old_focus_window=focus_window;
+   if(w==0)
+      return;
+   focus_window=0;
+   last_top_level_focus_window=0;
+   mode=GetKbdState(w);
+   if(mode==-1)
+      mode=0;
+   NewMode=(Mode&~SAVE_MASK)|mode;
+
+   NewMode&=~FOR_ONE;
+   FixNewMode();
+   SwitchKeyboard(NewMode);
+
+   if(prop)
+      XFree(prop);
+
+   focus_window=old_focus_window;
+   last_top_level_focus_window=old_last_top_level_focus_window;
+}
 static void InitCapsLockEmu()
 {
    KeyCode  *lock;
@@ -442,20 +615,33 @@ void load_map_delayed(const char *map)
 }
 
 #if TK!=TK_NONE
-void  RaiseButton()
+static void RaiseButton()
 {
+   if(AppData.wm_icon || AppData.wmaker_icon || top_level==0)
+      return;
    XRaiseWindow(disp,XtWindow(top_level));
 }
 
-void  RaiseButtonOnTimeout(XtPointer closure,XtIntervalId *id)
+static XtIntervalId raise_tmout=-1;
+static void RaiseButtonOnTimeout(XtPointer closure,XtIntervalId *id)
 {
    (void)closure; (void)id;
    raise_tmout=-1;
    RaiseButton();
 }
 
+static void SetTitle(int n)
+{
+   const char *title=(n==0 ? AppData.title0 : AppData.title1);
+   if(title==0)
+      title="xrus";
+   XtSetArg(args[0],XtNtitle,title);
+   XtSetArg(args[1],XtNiconName,title);
+   XtSetValues(top_level,args,2);
+}
+
 /* shows proper indicator button */
-void  ShowSwitchButton(void)
+static void ShowSwitchButton(void)
 {
    int   to_manage=-1;
 
@@ -479,6 +665,8 @@ void  ShowSwitchButton(void)
 
       XtManageChild(switch_button[to_manage]);
 
+      SetTitle(to_manage);
+
       RaiseButton();
    }
 }
@@ -513,12 +701,25 @@ void  VisibilityChange(Widget w,XtPointer closure,XEvent *ev,Boolean *cont)
                         (XtTimerCallbackProc)RaiseButtonOnTimeout,NULL);
    }
 }
-#endif /* TK!=TK_NONE */
 
-void  PerformSwitch(void)
+XtIntervalId pushed_tmout=-1;
+static void drop_pushed()
 {
-   SwitchKeyboard(Mode^MODE);
+   pushed_tmout=-1;
+   pushed=False;
 }
+static void PerformSwitch(void)
+{
+   NewMode=(Mode^MODE)&~FOR_ONE;
+   FixNewMode();
+   SwitchKeyboard(NewMode);
+   pushed=True;
+   if(pushed_tmout!=-1)
+      XtRemoveTimeOut(pushed_tmout);
+   pushed_tmout=XtAppAddTimeOut(app_context,2000,
+                        (XtTimerCallbackProc)drop_pushed,NULL);
+}
+#endif /* TK!=TK_NONE */
 
 /* a window is in cache => we have requested all needed input from the window */
 /* returns a boolean indicating the window was in cache
@@ -563,125 +764,124 @@ void  DelWindow(Window w)
    }
 }
 
-int   window_is_top_level(Window w)
-{
-   XGetWindowProperty(disp,w,XA_WM_CLASS,0L,0L,0,XA_STRING,
-         &type_ret,&format_ret,&nitems_ret,&bytes_after_ret,&prop);
-   return(type_ret!=None);
-}
+static void AddWindow(Window w,int tmout,int max_depth);
 
-int   is_wm_window(Window w)
-{
-   Window   root1,parent,*children;
-   unsigned children_num,i;
-   int   is=0;
-
-   if(XQueryTree(disp,w,&root1,&parent,&children,&children_num))
-   {
-      is|=(parent==root1);
-      for(i=0; !is && i<children_num; i++)
-      {
-         is|=window_is_top_level(children[i]);
-      }
-      XFree(children);
-   }
-   return is;
-}
-
-void  AddWindow(Window w,int tmout);
-
-void  AddWindowOnTimeout(XtPointer closure,XtIntervalId *id)
+static void AddWindowOnTimeout(XtPointer closure,XtIntervalId *id)
 {
    (void)id;
    DelWindow((Window)closure);
-   AddWindow((Window)closure,False);
+   AddWindow((Window)closure,False,1000000);
 }
 
-void  AddWindow(Window w,int tmout)
+static void XrusCheck(Window w)
+{
+   int class_len=strlen(AppClass);
+   XTextProperty prop;
+   unsigned char *zero;
+
+   if(!xrus_check)
+      return;
+
+   if(!XGetTextProperty(disp,w,&prop,XA_WM_CLASS))
+      return;
+
+   if(prop.format!=8)
+      return;
+
+   zero=memchr(prop.value,0,prop.nitems);
+   if(zero && zero+1+class_len<=prop.value+prop.nitems
+   && !memcmp(zero+1,AppClass,class_len)
+   && (zero+1+class_len==prop.value+prop.nitems
+       || zero[1+class_len]==0))
+   {
+      fprintf(stderr,"%s: xrus already runs on the display\n",program);
+      exit(1);
+   }
+   XFree(prop.value);
+}
+
+static void AddWindow(Window w,int tmout,int max_depth)
 {
    XWindowAttributes wa;
    long mask;
-
-   if(AddToWindowCache(w)) /* if the window was not in cache */
-   {
-      if(!XtWindowToWidget(disp,w))
-      {
-         /* if it is not our own window, unselect any events */
-         XSelectInput(disp,w,0);
-      }
-
-      if(!XGetWindowAttributes(disp,w,&wa))
-      {
-         DelWindow(w);
-         return;
-      }
-
-      mask=wa.your_event_mask;
-
-      /* we want to be notified of subwindows creation/destruction etc */
-      mask|=SubstructureNotifyMask;
-
-      /* we request events only if
-         1. the window already requested this kind of events
-         2. the window is blocking the events from propagation
-         3. this window is wm's one or top level one
-            (for the case the application itself doesn't want these events)
-      */
-      mask|=(wa.all_event_masks|wa.do_not_propagate_mask)
-            &(KeyPressMask|KeyReleaseMask|PointerMotionMask);
-
-      /* we always want both Press and Release events */
-      if(mask&(KeyPressMask|KeyReleaseMask))
-         mask|=(KeyPressMask|KeyReleaseMask);
-
-      /* third condition */
-      if((mask&(KeyPressMask|KeyReleaseMask))
-             !=(KeyPressMask|KeyReleaseMask)
-      && (w==wa.root || window_is_top_level(w) || is_wm_window(w)))
-         mask|=(KeyPressMask|KeyReleaseMask);
-      else
-         DelWindow(w);  /* we'll check input mask later */
-
-      if(w==wa.root)
-	 mask|=PointerMotionMask;   /* always want it for root window */
-
-      XSelectInput(disp,w,mask);
-
-      AddChildren(w);
-
-      if(tmout)
-      {
-         XtAppAddTimeOut(app_context,AppData.recheckTime,AddWindowOnTimeout,
-                         (XtPointer)w);
-      }
-   }
-}
-
-void  AddChildren(Window w)
-{
-   Window   root,parent,*children;
+   Window root1,parent,*children;
    unsigned children_num,i;
+   Bool all_events_allowed=False;
 
-   if(xrus_check)
+   if(max_depth<1)
+      return;
+
+   if(!AddToWindowCache(w)) /* if the window is in cache, skip it */
+      return;
+
+   XrusCheck(w);
+
+   if(!XQueryTree(disp,w,&root1,&parent,&children,&children_num))
    {
-      if(XGetWindowProperty(disp,w,XA_WM_CLASS,0L,256L,0,XA_STRING,
-            &type_ret,&format_ret,&nitems_ret,&bytes_after_ret,&prop)
-         ==Success && type_ret!=None)
-      {
-         if(!strcmp((char*)(prop+strlen((char*)prop)+1),AppClass))
-         {
-            fprintf(stderr,"%s: xrus already runs on the display\n",program);
-            exit(1);
-         }
-         XtFree(prop);
-      }
+      DelWindow(w);
+      return;
    }
 
-   if(XQueryTree(disp,w,&root,&parent,&children,&children_num))
+   if(w==focus_window)
+      all_events_allowed=True;
+
+   if(!all_events_allowed && !window_is_ours(w))
    {
-      for(i=0; i<children_num; i++)
-         AddWindow(children[i],False);
-      XFree(children);
+      /* if it is not our own window, unselect any events,
+         so that all_event_masks has only event bits of other apps. */
+      /* ...unless it is possible to select any events. */
+      XSelectInput(disp,w,0);
+   }
+
+   if(!XGetWindowAttributes(disp,w,&wa))
+   {
+      DelWindow(w);
+      return;
+   }
+
+   mask=wa.your_event_mask;
+
+   /* we want to be notified of subwindows creation/destruction etc */
+   mask|=SubstructureNotifyMask;
+
+   /* keep track of focus window */
+   mask|=FocusChangeMask;
+
+   /* we request events only if
+      1. the window owner has already requested this kind of events
+      2. the window is blocking the events from propagation
+      3. parent window belongs to a different client.
+   */
+   mask|=(wa.all_event_masks|wa.do_not_propagate_mask)
+         &(KeyPressMask|KeyReleaseMask|PointerMotionMask);
+
+   if(all_events_allowed)
+      mask|=(KeyPressMask|KeyReleaseMask|PointerMotionMask|PropertyChangeMask);
+
+   /* we always want both Press and Release events */
+   if(mask&(KeyPressMask|KeyReleaseMask))
+      mask|=(KeyPressMask|KeyReleaseMask);
+
+   if(!(mask&KeyPressMask))
+      DelWindow(w);  /* we'll check input mask later */
+
+   XSelectInput(disp,w,mask);
+
+   for(i=0; i<children_num; i++)
+   {
+      AddWindow(children[i],False,max_depth-1);
+      if(w==focus_window && window_is_top_level(children[i]))
+      {
+         last_top_level_focus_window=children[i];
+         last_top_level_revert_to=revert_to;
+      }
+   }
+   XFree(children);
+
+   if(tmout)
+   {
+      XtAppAddTimeOut(app_context,AppData.recheckTime,AddWindowOnTimeout,
+                      (XtPointer)w);
    }
 }
 
@@ -762,12 +962,59 @@ void  CheckKeymap()
    first_time=0;
 }
 
+static Window set_focus;
+static int set_revert_to;
+XtIntervalId focus_tmout=-1;
+static void SetFocusOnTimeout()
+{
+   if(set_focus && pushed)
+      XSetInputFocus(disp,set_focus,set_revert_to,CurrentTime);
+   focus_tmout=-1;
+}
+static void NewFocus(Window w,int rt)
+{
+   if(w==focus_window)
+      return;
+   if((top_level && w==XtWindow(top_level)) || w==our_host_window)
+   {
+      if(focus_tmout!=-1)
+         XtRemoveTimeOut(focus_tmout);
+      set_focus=last_top_level_focus_window;
+      set_revert_to=last_top_level_revert_to;
+      focus_tmout=XtAppAddTimeOut(app_context,200,
+                        (XtTimerCallbackProc)SetFocusOnTimeout,NULL);
+      return;
+   }
+   if(window_is_top_level(w))
+   {
+      last_top_level_focus_window=w;
+      last_top_level_revert_to=rt;
+      if(AppData.per_window_state)
+         SwitchKeyboardForWindow(w);
+   }
+   focus_window=w;
+   revert_to=rt;
+   DelWindow(w);
+   AddWindow(w,False,1);
+}
+
+static void QueryFocus()
+{
+   Window w=0;
+   int revert_to=0;
+   XGetInputFocus(disp,&w,&revert_to);
+   if(w)
+      NewFocus(w,revert_to);
+}
+
 static
 void  MappingNotifyTimeoutHandler(XtPointer closure,XtIntervalId *id)
 {
    (void)closure; (void)id;
    MappingNotifyTimeout=-1;
    CheckKeymap();
+   if(focus_window==0)
+      QueryFocus();
 }
 
 void  MainLoop(void)
@@ -796,11 +1043,14 @@ void  MainLoop(void)
       switch(ev.type)
       {
       case(MapNotify):
-         AddWindow(ev.xmap.window,True);
+         AddWindow(ev.xmap.window,True,1000000);
+         QueryFocus();
          break;
       case(ReparentNotify):
          /* the structure could change under us - re-add the window */
-         AddWindow(ev.xreparent.window,True);
+         AddWindow(ev.xreparent.window,True,1000000);
+         if(ev.xreparent.window==XtWindow(top_level))
+            our_host_window=ev.xreparent.parent;
          break;
       case(DestroyNotify):
          DelWindow(ev.xdestroywindow.window);
@@ -872,6 +1122,26 @@ void  MainLoop(void)
             }
          }
          break;
+      case(FocusIn):
+         QueryFocus();
+         break;
+      case(FocusOut):
+      {
+         Window old=focus_window;
+         focus_window=0;
+         DelWindow(ev.xany.window);
+         AddWindow(ev.xany.window,False,1);
+         focus_window=old;
+         break;
+      }
+      case(PropertyNotify):
+         if(AppData.per_window_state && ev.xproperty.atom==XA_WM_NAME
+         && ev.xproperty.state==PropertyNewValue)
+         {
+            XtAppAddTimeOut(app_context,1000,
+                        (XtTimerCallbackProc)SetTitleIndicatorOnTimeout,
+                        (XtPointer)ev.xproperty.window);
+         }
       }
    }
 }
@@ -1071,7 +1341,7 @@ int   main(int argc,char **argv)
       {
          printf("Xrus(keyboard switcher) version " VERSION "\n"
                 "\n"
-                "Copyright (c) 1995-98 Alexander V. Lukyanov (lav@yars.free.net)\n"
+                "Copyright (c) 1995-2000 Alexander V. Lukyanov (lav@yars.free.net)\n"
                 "This is free software that gives you freedom to use, modify and distribute\n"
                 "it under certain conditions, see COPYING (GNU GPL) for details.\n"
                 "There is ABSOLUTELY NO WARRANTY, use at your own risk.\n");
@@ -1094,6 +1364,7 @@ int   main(int argc,char **argv)
                 " -nofork                don't fork to background\n"
                 " -wmaker                display in Window Maker appicon\n"
                 " -wmicon                display in window manager icon\n"
+                " -perwindow             remember state for each window\n"
                 "\n",program);
          exit(0);
       }
@@ -1110,6 +1381,8 @@ int   main(int argc,char **argv)
       fprintf(stderr,"%s: cannot open display\n",program);
       exit(1);
    }
+   client_id_mask=~(disp->resource_mask);
+
    if(argc>1 && argv[1][0]=='-')
    {
       fprintf(stderr,"%s: invalid option -- %s\n",program,argv[1]);
@@ -1130,15 +1403,16 @@ int   main(int argc,char **argv)
 
    old_error_handler=XSetErrorHandler(X_error_handler);
 
-   /* add all windows located on the display to the pool
-      and check for other xrus presence */
+   /* select events from all windows located on the display
+      and check for another xrus instance */
    for(scr=0; scr<ScreenCount(disp); scr++)
-      AddWindow(RootWindowOfScreen(ScreenOfDisplay(disp,scr)),True);
+      AddWindow(RootWindowOfScreen(ScreenOfDisplay(disp,scr)),True,1000000);
    xrus_check=0;
 
    StartArgs();
 #if TK==TK_MOTIF
    AddArg(XmNmwmDecorations,MWM_DECOR_BORDER);
+   AddArg(XmNmwmFunctions,MWM_FUNC_MOVE|MWM_FUNC_CLOSE);
 #elif TK==TK_NONE
    AddArg(XtNwidth,10); /* no toolkit - set size to avoid 0x0 size */
    AddArg(XtNheight,10);
@@ -1152,6 +1426,8 @@ int   main(int argc,char **argv)
 
    XtGetApplicationResources(top_level,&AppData,resources,XtNumber(resources),NULL,0);
 
+   SetTitle(0);
+
    /* replace pointer to constant string with pointer to our buffers */
    strcpy(LockerData,AppData.locker?AppData.locker:DefaultLocker);
    AppData.locker=LockerData;
@@ -1221,7 +1497,7 @@ int   main(int argc,char **argv)
 #if TK!=TK_NONE
    XtAddEventHandler(top_level,StructureNotifyMask,False,UnmapHandler,NULL);
 
-#if TK==TK_MOTIF
+# if TK==TK_MOTIF
    StartArgs();
    AddArg(XmNresizable,False);
    form_w=XtCreateManagedWidget("form",xmFormWidgetClass,top_level,args,count);
@@ -1231,7 +1507,7 @@ int   main(int argc,char **argv)
 
    XtAddCallback(switch_button[0],XmNactivateCallback,(XtCallbackProc)PerformSwitch,NULL);
    XtAddCallback(switch_button[1],XmNactivateCallback,(XtCallbackProc)PerformSwitch,NULL);
-#else
+# elif TK==TK_XAW
    form_w=XtCreateManagedWidget("form",boxWidgetClass,top_level,NULL,0);
 
    switch_button[0]=XtCreateWidget("modeButton0",commandWidgetClass,form_w,NULL,0);
@@ -1239,7 +1515,9 @@ int   main(int argc,char **argv)
 
    XtAddCallback(switch_button[0],XtNcallback,(XtCallbackProc)PerformSwitch,NULL);
    XtAddCallback(switch_button[1],XtNcallback,(XtCallbackProc)PerformSwitch,NULL);
-#endif
+# else
+#  error unknown toolkit
+# endif
 
    XtManageChild(switch_button[(Mode&MODE)==RUS]);
 
@@ -1247,7 +1525,19 @@ int   main(int argc,char **argv)
    XtAddEventHandler(switch_button[1],VisibilityChangeMask,False,VisibilityChange,NULL);
 
    XrusMenuCreate();
-#endif
+
+#endif /* !TK_NONE */
+
+   kbd_state_atom=XInternAtom(disp,"KBD_STATE",False);
+
+   XtRealizeWidget(top_level);
+
+   KeyboardStateInit();
+   InitCapsLockEmu();
+
+#if TK!=TK_NONE
+   /* Set various window properties. */
+   w=XtWindow(top_level);
 
    ol_decor_del_atom=XInternAtom(disp,"_OL_DECOR_DEL",False);
    ol_decor_del[0]=XInternAtom(disp,"_OL_DECOR_HEADER",False);
@@ -1256,19 +1546,66 @@ int   main(int argc,char **argv)
    wm_client_leader=XInternAtom(disp,"WM_CLIENT_LEADER",False);
    wm_delete_window=XInternAtom(disp,"WM_DELETE_WINDOW",False);
 
-   XtRealizeWidget(top_level);
-
-   XChangeProperty(disp,XtWindow(top_level),ol_decor_del_atom,XA_ATOM,32,
+   XChangeProperty(disp,w,ol_decor_del_atom,XA_ATOM,32,
          PropModeReplace,(void*)ol_decor_del,2);
 
-   w=XtWindow(top_level);
-   XChangeProperty(disp,XtWindow(top_level),wm_client_leader,XA_WINDOW,32,
+   XChangeProperty(disp,w,wm_client_leader,XA_WINDOW,32,
          PropModeReplace,(void*)&w,1);
 
-#if TK!=TK_MOTIF && TK!=TK_NONE
+#define WinStateAllWorkspaces  (1 << 0)   /* appears on all workspaces */
+#define WinStateMinimized      (1 << 1)   /* to iconbox,taskbar,... */
+#define WinStateMaximizedVert  (1 << 2)   /* maximized vertically */
+#define WinStateMaximizedHoriz (1 << 3)   /* maximized horizontally */
+#define WinStateHidden         (1 << 4)   /* not on taskbar if any, but still accessible */
+#define WinStateRollup         (1 << 5)   /* only titlebar visible */
+#define WinStateFixedPosition  (1 << 10)  /* fixed position on virtual desktop*/
+#define WinStateArrangeIgnore  (1 << 11)  /* ignore for auto arranging */
+#define WinStateWithdrawn      (1 << 31)  /* managed, but not available to user */
+
+if(AppData.occupyAllDesks)
+{  /* make the window global for all desktops */
+   Atom win_state,sgi_desks_hints,sgi_desks_always_global;
+   Atom kwm_win_sticky;
+   long value;
+   /* for gnome */
+   static long state[2]={WinStateAllWorkspaces|WinStateArrangeIgnore,63};
+   win_state=XInternAtom(disp,"_WIN_STATE",False);
+   XChangeProperty(disp,w,win_state,XA_CARDINAL,32,
+      PropModeReplace,(void*)state,2);
+   /* for SGI */
+   sgi_desks_hints=XInternAtom(disp,"_SGI_DESKS_HINTS",False);
+   sgi_desks_always_global=XInternAtom(disp,"_SGI_DESKS_ALWAYS_GLOBAL",False);
+   value=sgi_desks_always_global;
+   XChangeProperty(disp,w,sgi_desks_hints,XA_ATOM,32,
+      PropModeReplace,(void*)&value,1);
+   /* for KDE */
+   value=1;
+   kwm_win_sticky=XInternAtom(disp,"KWM_WIN_STICKY",False);
+   XChangeProperty(disp,w,kwm_win_sticky,kwm_win_sticky,32,
+      PropModeReplace,(void*)&value,1);
+}
+
+#define WinHintsSkipFocus      (1 << 0)
+#define WinHintsSkipWindowMenu (1 << 1)
+#define WinHintsSkipTaskBar    (1 << 2)
+#define WinHintsGroupTransient (1 << 3)
+#define WinHintsDockHorizontal (1 << 6)   /* docked horizontally */
+
+{
+   Atom win_hints=XInternAtom(disp,"_WIN_HINTS",False);
+   long value[2];
+   value[0]=WinHintsSkipFocus|WinHintsSkipWindowMenu;
+   if(!AppData.wm_icon)
+      value[0]|=WinHintsSkipTaskBar;
+   value[1]=127;
+   XChangeProperty(disp,w,win_hints,XA_CARDINAL,32,
+      PropModeReplace,(void*)value,2);
+}
+
+#if TK!=TK_MOTIF
 {
    /* we have to emulate motif hints as we don't have Motif */
-   static long motif_hints[]={0x2,0xffffffff,0x2,0xffffffff,0x39250};
+   static long motif_hints[]={0x3, 0x24, 0x2, 0xffffffff, 0};
    Atom motif_hints_atom=XInternAtom(disp,"_MOTIF_WM_HINTS",False);
    XChangeProperty(disp,XtWindow(top_level),motif_hints_atom,motif_hints_atom,32,
          PropModeReplace,(void*)motif_hints,XtNumber(motif_hints));
@@ -1278,10 +1615,6 @@ int   main(int argc,char **argv)
 }
 #endif
 
-   KeyboardStateInit();
-   InitCapsLockEmu();
-
-#if TK!=TK_NONE
    if(AppData.adjustModeButtons)
    {
       Dimension w0,h0,w1,h1;
